#!/usr/bin/env python

import math
import sys
import gc
import rospy
import re
import array
import numpy as np
from datetime import datetime
from geometry_msgs.msg import Quaternion
from tf.transformations import euler_from_quaternion, quaternion_from_euler
from nav_msgs.msg import Odometry
from gazebo_msgs.srv import GetModelState
from gazebo_msgs.srv import GetWorldProperties

BotDist = float(sys.argv[1])#Erstes Argumnet
SafeDist = float(sys.argv[2])#Zweites argument
DestinationX = float(sys.argv[3])#Ziellinie
Timeout = float(sys.argv[4])#timeout
yawBot = 0.0
WriteFlag = 0
logtime = "location_monitor_log.txt"
f = open(logtime, "w+")


# y
# ^
# |
# |  -------          Box muss im rechtem Winkel zur x-Achse sein!
# |  |     |
# |  |     |
# |  -------
# |
# ---------------->x


class Block:
    def __init__(self, name, relative_entity_name, radius):
        self._name = name
        self._relative_entity_name = relative_entity_name
        self._radius = radius


class Detection:
    _blockListDict = {}
    def show_gazebo_models(self, botx, boty):
     global Timeout
     global DestinationX
     global f
     global WriteFlag
     get_world_properties = rospy.ServiceProxy('/gazebo/get_world_properties', GetWorldProperties)
     resp1 = get_world_properties()
     if resp1.sim_time >= Timeout:
         print("Simulation Timeout!")
         if WriteFlag == 0:
            WriteFlag = 1
            f.write("[T:" + str(resp1.sim_time) + "] " + "Simulation Timeout!" + "\n")
            f.close
     if WriteFlag == 0:
         f.write("[T:" + str(resp1.sim_time) + "] " + "Bot X: " + str(botx) + "\n")
         f.write("[T:" + str(resp1.sim_time) + "] " + "Bot Y: " + str(boty) + "\n")
     if botx >= DestinationX:
         print("Bot reached Destination!")
         if WriteFlag == 0:
            WriteFlag = 1
            f.write("[T:" + str(resp1.sim_time) + "] " + "Bot reached Destination!" + "\n")
            f.close
     if botx > 0 and boty > 0:
        i = 0
        for model in resp1.model_names:
            if ("ball" in str(model)) or ("cylinder" in str(model)) or ("box" in str(model) or ("sphere" in str(model))):
                if("R_" in str(model)):
                    keyword = "R_"
                    search = str(model)
                    afsearch = search.partition(keyword)
                    matches = re.findall(r"[-+]?\d*\.\d+|\d+", afsearch[2])
                else:
                    keyword = "S_"
                    search = str(model)
                    afsearch = search.partition(keyword)
                    matches = re.findall(r"[-+]?\d*\.\d+|\d+", afsearch[2])
                rad = 0
                if matches:
                    if "box" in str(model):
                        rad = (matches[0])
                        rad = float(rad) / 2.0
                    else:
                        rad = (matches[0])
                s = 'block_' + str(i)
                self._blockListDict.update({s: Block(model, 'link', rad)})
                i = i + 1
        try:
            model_coordinates = rospy.ServiceProxy('/gazebo/get_model_state', GetModelState)
            for block in self._blockListDict.itervalues():
                blockName = str(block._name)
                if blockName in resp1.model_names:
                    resp_coordinates = model_coordinates(blockName, block._relative_entity_name)
                    l_x = resp_coordinates.pose.position.x
                    l_y = resp_coordinates.pose.position.y
                    print(blockName)
                    print("X: " + str(l_x))
                    print("Y: " + str(l_y))
                    if WriteFlag == 0:
                        f.write(str(blockName) + "\n")
                        f.write("[T:" + str(resp1.sim_time) + "] " + str(blockName) + " X: " + str(l_x) + "\n")
                        f.write("[T:" + str(resp1.sim_time) + "] " + str(blockName) + " Y: " + str(l_y) + "\n")
                    if block._radius != 0:

                        if "box" in str(blockName):
                            box_orient_list = [resp_coordinates.pose.orientation.x,resp_coordinates.pose.orientation.y,resp_coordinates.pose.orientation.z,resp_coordinates.pose.orientation.w]
                            (roll, pitch, yaw) = euler_from_quaternion(box_orient_list)
                            yaw_degrees = yaw * 180.0 / math.pi
                            if yaw_degrees < 0:
                                yaw_degrees += 360.0
                            boxkdetection(float(block._radius), botx, boty, l_x, l_y, yaw_degrees, resp1.sim_time)
                        else: #if "sphere" , "cylinder" , "ball" etc.
                            objectWithRadiusDetection(float(block._radius), botx, boty, l_x, l_y, str(resp1.sim_time))


                    else:
                        print("Dist: " + str(dist))
                        if WriteFlag == 0:
                            f.write("[T:" + str(resp1.sim_time) + "] " + "Dist: " + str(dist) + "\n")

        except rospy.ServiceException as e:
            rospy.loginfo("Get Model State service call failed:  {0}".format(e))
     else:
         rospy.loginfo("Bot out of tracking Range!((x,y)<0)")
         if WriteFlag == 0:
             f.write("[T:" + str(resp1.sim_time) + "] " + "Bot out of tracking Range!((x,y)<0)" + "\n")
             f.close

def distance(x1, y1, x2, y2):
    dist = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
    return dist

def objectWithRadiusDetection(radius, botx, boty, l_x, l_y, simtime):
    global WriteFlag
    global f
    global SafeDist
    dist = distance(botx, boty, l_x, l_y)
    safedist = radius + SafeDist  # Radius/Size + Bot front Rad + safe dist
    print("Safedist: " + str(safedist))
    if WriteFlag == 0:
        f.write("Safedist.: " + str(safedist) + "\n")
    print("Distance: " + str(dist))
    if WriteFlag == 0:
        f.write("[T:" + simtime + "] " + "Dist. To Collision Point: " + str(dist) + "\n")
    if dist < safedist:
        rospy.loginfo("Collision!")
        if WriteFlag == 0:
            f.write("[T:" + simtime + "] " + "Collision!")
            f.close
            WriteFlag = 1


def boxkdetection(size, botx, boty, l_x, l_y, yaw_degrees, simtime):
    global f
    global WriteFlag
    if (size * 2.0) > 5.0:
        print("Box to big, > 5m")
        if WriteFlag == 0:
            f.write("[T:" + str(simtime) + "] " + "Box Invalid (size > 5m)" + "\n")
            f.close
    else:
        # print("SIZEBOX" + str(size))
        safedist =  SafeDist
        #print(str(safedist))
        if WriteFlag == 0:
            f.write("Safedist.: " + str(safedist) + "\n")
        xR = l_x + size * math.cos(math.radians(yaw_degrees))
        yL = l_y + size * math.sin(math.radians(yaw_degrees))

        #XUr = xR * math.cos(math.radians(45)) - yL * math.sin(math.radians(45))
        #yUr = xR * math.sin(math.radians(45)) + yL * math.cos(math.radians(45))
        #dist1 = distance(botx, boty, XUr, yUr)

        #XUl = xR * math.cos(math.radians(135)) - yL * math.sin(math.radians(135))
        #yUl = xR * math.sin(math.radians(135)) + yL * math.cos(math.radians(135))
        #dist2 = distance(botx, boty, XUl, yUl)

        #XDl = xR * math.cos(math.radians(225)) - yL * math.sin(math.radians(225))
        #yDl = xR * math.sin(math.radians(225)) + yL * math.cos(math.radians(225))
        #dist3 = distance(botx, boty, XDl, yDl)

        #XDr = xR * math.cos(math.radians(315)) - yL * math.sin(math.radians(315))
        #yDr = xR * math.sin(math.radians(315)) + yL * math.cos(math.radians(315))
        #dist4 = distance(botx, boty, XDr, yDr)

        #print(str(xR) + " " + str(yL))#korrekt

        lineA = ((botx, boty), (l_x, l_y))
        lineB = ((l_x, l_y), (xR, yL))

        angle = angle_between(lineA, lineB)
        disangle = angle
        angle1 = angle
        if angle > 45:
            disangle = 90 - angle;
        #print("Angle: " + str(angle))#korrekt
        dis = (size) / math.cos(math.radians(disangle))
        # print("Dist from center do edge: " + str(dis))

        #f dist1 < dist2 and dist1 < dist3 and dist1 < dist4:
        #    angle2 = angle
        #if dist2 < dist1 and dist2 < dist3 and dist2 < dist4:
        #    angle2 = 180 - angle
        #if dist3 < dist1 and dist3 < dist2 and dist3 < dist4:
        #    angle2 = 180 + angle
        #if dist4 < dist1 and dist4 < dist2 and dist4 < dist3:
        #    angle2 = 360 - angle

        if botx < l_x and boty > l_y:
            angle1 = 180 - angle
        if botx < l_x and boty < l_y:
            angle1 = 180 + angle
        if botx > l_x and boty < l_y:
            angle1 = 360 - angle

        #print("Angle1: " + str(angle2))
        dynX = l_x + dis * math.cos(math.radians(angle1))
        dynY = l_y + dis * math.sin(math.radians(angle1))

        #dynX = l_x + dis * math.cos(math.radians(angle2))
        #dynY = l_y + dis * math.sin(math.radians(angle2))

        #print("DynX : " + str(dynX))
        #print("DynY : " + str(dynY))
        distult = distance(botx, boty, dynX, dynY)
        if WriteFlag == 0:
            f.write("[T:" + str(simtime) + "] " + "Dist to collision point: " + str(distult) + "\n")
        print("Dist to collision point: " + str(distult))
        if distult < safedist:
            rospy.loginfo("Collision!")
            if WriteFlag == 0:
                f.write("[T:" + str(simtime) + "] " + "Collision!")
                f.close
                WriteFlag = 1


def slope(x1, y1, x2, y2):  # Line slope given two points:
    return (y2 - y1) / (x2 - x1)


def angle(s1, s2):
    return math.degrees(math.atan((s2 - s1) / (1 + (s2 * s1))))


def angle_between(lineA, lineB):
    slope1 = slope(lineA[0][0], lineA[0][1], lineA[1][0], lineA[1][1])
    slope2 = slope(lineB[0][0], lineB[0][1], lineB[1][0], lineB[1][1])

    ang = angle(slope1, slope2)
    if ang < 0:
        ang = ang * (-1.0)

    return ang

def get_rotation (msg):
    global yawBot
    orientation_q = msg.pose.pose.orientation
    orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
    (roll, pitch, yaw) = euler_from_quaternion (orientation_list)
    yaw_degrees = yaw * 180.0 /math.pi
    if yaw_degrees <0:
        yaw_degrees += 360.0
    #print("Yaw:" + str(yaw_degrees))
    yawBot = yaw_degrees

def callback(msg):
    global BotDist
    global yawBot
    get_rotation(msg)
    botx = msg.pose.pose.position.x
    boty = msg.pose.pose.position.y
    botx = botx + BotDist * math.cos(math.radians(yawBot))
    boty = boty + BotDist * math.sin(math.radians(yawBot))
    print("BotX: " + str(botx))
    print("BotY: " + str(boty))
    detect = Detection()
    detect.show_gazebo_models(botx, boty)
    gc.collect()



def main():
    rospy.init_node('location_monitor') #Node in gang bringen
    try:
        rospy.Subscriber("/odom", Odometry, callback) #Koordinaten des bots werden durch /odom geschickt und an callback gesendet
        gc.collect()
        rospy.spin()
    except rospy.ROSInterruptException:
        pass

if __name__ == '__main__':
    main()
